/* Internal MPM header.

   Copyright 2001, 2002, 2007 Kevin Ryde

   This file is part of MPM.

   MPM is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2, or (at your option) any later version.

   MPM is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MPM; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */


#include <stdlib.h>
#include "tests.h"



/* fields of an mpr_t */
#define RPTR(r)  ((r)->_mp_d)

/* fields of an mpm_t */
#define MZ(m)                ((m)->_mp_m)
#define MRSIZE(m)            ((m)->_mp_rsize)
#define MRALLOC(m)           ((m)->_mp_ralloc)
#define MPTR(m)              PTR(MZ(m))
#define MINV(m)              ((m)->_mp_inv)
#define MSMALLINV(m)         ((m)->_mp_smallinv)
#define MRMOD(m)             ((m)->_mp_rmod)
#define MR2MOD(m)            ((m)->_mp_r2mod)
#define MR3MOD(m)            ((m)->_mp_r3mod)
#define MNEGONE(m)           ((m)->_mp_negone)
#define MNORM(m)             ((m)->_mp_norm)
#define MNORM_LSHIFT(m)      ((m)->_mp_norm_lshift)
#define MNORM_RSHIFT(m)      ((m)->_mp_norm_rshift)
#define MNORM_RSHIFT_MASK(m) ((m)->_mp_norm_rshift_mask)
#define MUDIV_DH(m)          ((m)->_mp_norm_dh)
#define MUDIV_DH_INV(m)      ((m)->_mp_norm_dh_inv)
#define MUDIV_DL(m)          ((m)->_mp_norm_dl)
#define MDIV2CEIL(m)         ((m)->_mp_div2ceil)
#define MBITS(m)             ((m)->_mp_mbits)


#define ASSERT_ALWAYS_MPM(m)            \
  do {                                  \
    ASSERT (MRSIZE(m) >= 1);            \
    ASSERT (MRALLOC(m) >= 1);           \
    ASSERT (MRALLOC(m) >= MRSIZE(m));   \
    ASSERT_MPZ (MZ(m));                 \
  } while (0)

#if WANT_ASSERT
#define ASSERT_MPM(m)    ASSERT_ALWAYS_MPM (m)
#else
#define ASSERT_MPM(r,m)  do { } while (0)
#endif

#if WANT_ASSERT
#define ASSERT_MPR_MOD(r,m)                             \
  do {                                                  \
    ASSERT (MRSIZE(m) >= 1);                            \
    ASSERT (SIZ(MZ(m)) == MRSIZE(m));                   \
    ASSERT_MPN (RPTR(r), MRSIZE(m));                    \
    ASSERT (mpn_cmp (RPTR(r), MPTR(m), MRSIZE(m)) < 0); \
  } while (0)
#else
#define ASSERT_MPR_MOD(r,m)  do { } while (0)
#endif

#if WANT_ASSERT
#define ASSERT_MPZ(z)   MPZ_CHECK_FORMAT(z)
#else
#define ASSERT_MPZ(z)   do {} while (0)
#endif

#define MPR_TMP_INIT(r, m)                      \
  do {                                          \
    (r) = (mpr_t) TMP_ALLOC_LIMBS (MRALLOC(m)); \
  } while (0)


extern mp_size_t  redc_big_threshold;

#define mpn_modlimbs_invert __MPN(modlimbs_invert)
void mpn_modlimbs_invert __GMP_PROTO ((mp_ptr, mp_srcptr, mp_size_t));

void mpm_init_redc __GMP_PROTO ((mpm_ptr, mpz_srcptr));
void mpm_init_one __GMP_PROTO ((mpm_ptr, mp_limb_t));
void mpm_mod__clear __GMP_PROTO ((mpm_ptr));
void mpm_mod__init __GMP_PROTO ((mpm_ptr, mpz_srcptr));

void mpr_trace __GMP_PROTO ((const char *, mpr_srcptr, mpm_srcptr));

void mpr_2exp__mul __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpr_srcptr, mpm_srcptr));

void mpr_2expsub1__mul__shift __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpr_srcptr, mpm_srcptr));
void mpr_2expsub1__mul__noshift __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpr_srcptr, mpm_srcptr));

void mpr_div__mul __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpr_srcptr, mpm_srcptr));
void mpr_div__set_ui (mpr_ptr w, unsigned long x, mpm_srcptr m);

void mpr_gen__clear __GMP_PROTO ((mpr_ptr, mpm_srcptr));
void mpr_gen__div__invert_mul __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpr_srcptr, mpm_srcptr));
int mpr_gen__equal __GMP_PROTO ((mpr_srcptr, mpr_srcptr, mpm_srcptr));
int mpr_gen__equal_negone __GMP_PROTO ((mpr_srcptr, mpm_srcptr));
int mpr_gen__equal_zero __GMP_PROTO ((mpr_srcptr, mpm_srcptr));
void mpr_gen__gcd__odd __GMP_PROTO ((mpz_ptr, mpr_srcptr, mpm_srcptr));
unsigned long mpr_gen__get_ui__get_z __GMP_PROTO ((mpr_srcptr, mpm_srcptr));
int mpr_gen__invert __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpm_srcptr));
int mpr_gen__jacobi (mpr_srcptr, mpm_srcptr);
mpr_ptr mpr_gen__new __GMP_PROTO ((mpm_srcptr));
void mpr_gen__set __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpm_srcptr));
void mpr_gen__set_ui__fits __GMP_PROTO ((mpr_ptr, unsigned long, mpm_srcptr));
void mpr_gen__swap __GMP_PROTO ((mpr_ptr, mpr_ptr, mpm_srcptr));

void mpr_mod__add __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpr_srcptr, mpm_srcptr));
void mpr_mod__addmul_ui __GMP_PROTO ((mpr_ptr, mpr_srcptr, unsigned long, mpm_srcptr));
void mpr_mod__div_2exp __GMP_PROTO ((mpr_ptr, mpr_srcptr, unsigned long, mpm_srcptr));
void mpr_mod__mul_2exp __GMP_PROTO ((mpr_ptr, mpr_srcptr, unsigned long, mpm_srcptr));
void mpr_mod__mul_si __GMP_PROTO ((mpr_ptr, mpr_srcptr, long, mpm_srcptr));
void mpr_mod__mul_ui __GMP_PROTO ((mpr_ptr, mpr_srcptr, unsigned long, mpm_srcptr));
void mpr_mod__neg __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpm_srcptr));
void mpr_mod__reduce_1 __GMP_PROTO ((mp_limb_t, mp_ptr wp, mpm_srcptr));
void mpr_mod__set_z __GMP_PROTO ((mpr_ptr, mpz_srcptr, mpm_srcptr));
void mpr_mod__sub __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpr_srcptr, mpm_srcptr));
void mpr_mod__submul_ui __GMP_PROTO ((mpr_ptr, mpr_srcptr, unsigned long, mpm_srcptr));
void mpr_mod__urandomm __GMP_PROTO ((mpr_ptr, gmp_randstate_ptr, mpm_srcptr));

void mpr_one__mul __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpr_srcptr, mpm_srcptr));

void mpr_redc__mul_small __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpr_srcptr, mpm_srcptr));
void mpr_redc__mul_big __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpr_srcptr, mpm_srcptr));


#define ASSERT_MPR_GCD(w, m)            \
  do {                                  \
    ASSERT_MPZ (w);                     \
    ASSERT (mpz_cmp_ui (w, 1L) >= 0);   \
    ASSERT (mpz_cmp (w, MZ(m)) <= 0);   \
  } while (0)

#define ASSERT_MPR_GET_Z(z,m)           \
  do {                                  \
    ASSERT_MPZ (z);                     \
    ASSERT (mpz_cmp_ui (z, 0L) >= 0);   \
    ASSERT (mpz_cmp (z, MZ(m)) < 0);    \
  } while (0)

#define ASSERT_MPR_GET_UI(u,m)          \
  do {                                  \
    ASSERT (mpz_cmp_ui (MZ(m), u) > 0); \
  } while (0)

/* Execute the given action if {xp,size} < {yp,size}. */
#define MPN_CMP_LT_ACTION(xp, yp, size, action) \
  {                                             \
    mp_srcptr  __xp = (xp);                     \
    mp_srcptr  __yp = (yp);                     \
    mp_size_t  __i = (size);                    \
                                                \
    ASSERT ((size) >= 0);                       \
                                                \
    __i--;                                      \
    if (LIKELY (__i >= 0))                      \
      {                                         \
        mp_limb_t  __x, __y;                    \
        do                                      \
          {                                     \
            __x = __xp[__i];                    \
            __y = __yp[__i];                    \
            if (LIKELY (__x != __y))            \
              break;                            \
          }                                     \
        while (--__i >= 0);                     \
                                                \
        if (__x < __y)                          \
          { action; }                           \
      }                                         \
  }


/* Swap the contents of {xp,size} and {yp,size}.  */
#define MPN_SWAP(xp, yp, size)                  \
  do {                                          \
    mp_size_t  __i;                             \
    CRAY_Pragma ("_CRI ivdep");                 \
    for (__i = 0; __i < (size); __i++)          \
      MP_LIMB_T_SWAP ((xp)[__i], (yp)[__i]);    \
  } while (0)


/* declare an mpz_t variable "z" initialized to the long value "s" */
#if GMP_NAIL_BITS == 0
#define MPZ_FAKE_SI_VAR(z,s)                                    \
  mpz_t z;                                                      \
  mp_limb_t __zl = (unsigned long) ((s) >= 0 ? (s) : -(s));     \
  PTR(z) = &__zl;                                               \
  SIZ(z) = ((s) > 0 ? 1 : (s) == 0 ? 0 : -1);                   \
  ASSERT_CODE (ALLOC(z) = 1);                                   \
  ASSERT_MPZ (z);
#else
#define MPZ_FAKE_SI_VAR(z,s)                                            \
  MPZ_FAKE_UI_VAR (z, ((s) >= 0 ? (s) : - (unsigned long) (s)));        \
  if ((s) < 0)                                                          \
    SIZ(z) = -SIZ(z);                                                   \
  ASSERT_MPZ (z);
#endif

/* declare an mpz_t variable "z" initialized to the unsigned long value "u" */
#if GMP_NAIL_BITS == 0
#define MPZ_FAKE_UI_VAR(z,u)    \
  mpz_t     z;                  \
  mp_limb_t __zl = (u);         \
  PTR(z) = &__zl;               \
  SIZ(z) = (__zl != 0);         \
  ASSERT_CODE (ALLOC(z) = 1);   \
  ASSERT_MPZ (z);
#else
#define MPZ_FAKE_UI_VAR(z,u)                            \
  mpz_t     z;                                          \
  mp_limb_t __z[2];                                     \
  __z[0] = (u) & GMP_NUMB_MASK;                         \
  __z[1] = (u) >> GMP_NUMB_BITS;                        \
  SIZ(z) = (__z[1] != 0 ? 2 : __z[0] != 0 ? 1 : 0);     \
  PTR(z) = __z;                                         \
  ASSERT_CODE (ALLOC(z) = 2);                           \
  ASSERT_MPZ (z);
#endif

/* set the mpz_t variable z to be a fake of mpr_t r (under mpm_t m)
   the actual value depends on the residue representation within the mpm_t
   type */
#define MPZ_FAKE_R(z,r,m)               \
  MPZ_FAKE_N (z, RPTR(r), MRSIZE(m))

/* set the mpz_t variable z to be a fake pointing to {ptr,size}
   any leading zero limbs on {ptr,size} are stripped */
#define MPZ_FAKE_N(z, ptr, size)                                \
  do {                                                          \
    /* no cast so as to typecheck, then cast to hide const */   \
    mp_srcptr  __mpzfaken__srcptr = (ptr);                      \
    mp_ptr     __mpzfaken__ptr = (mp_ptr) __mpzfaken__srcptr;   \
    mp_size_t  __mpzfaken__size = (size);                       \
    MPN_NORMALIZE (__mpzfaken__ptr, __mpzfaken__size);          \
    PTR(z) = __mpzfaken__ptr;                                   \
    SIZ(z) = __mpzfaken__size;                                  \
    ASSERT_CODE (ALLOC(z) = (size));                            \
    ASSERT_MPZ (z);                                             \
  } while (0)

/* If {ptr,size} is all zeros, then execute "action".
   If {ptr,size} is not all zeros then do nothing.  */
#define MPN_NONZERO_P_ACTION(ptr,size,action)   \
  do {                                          \
    mp_size_t  __i;                             \
    for (__i = 0; __i < (size); __i++)          \
      if (LIKELY ((ptr)[__i] != 0))             \
        action;                                 \
  } while (0)

/* If {xp,xsize} is equal to {yp,ysize} then set "result" to 1.
   If they're not equal then set "result" to 0.  */
#define MPN_EQUAL(result, xp, yp, size) \
  do {                                  \
    mp_srcptr  __xp = (xp);             \
    mp_srcptr  __yp = (yp);             \
    mp_size_t  __size = (size);         \
    mp_size_t  __i;                     \
    (result) = 1;                       \
    for (__i = 0; __i < __size; __i++)  \
      if (__xp[__i] != __yp[__i])       \
        {                               \
          (result) = 0;                 \
          break;                        \
        }                               \
  } while (0)

/* If {xp,xsize} is equal to {yp,ysize} then do "return 1".
   If they're not equal do "return 0".  */
#define MPN_EQUAL_RETURN(xp,yp,size)    \
  do {                                  \
    mp_srcptr  __xp = (xp);             \
    mp_srcptr  __yp = (yp);             \
    mp_size_t  __size = (size);         \
    mp_size_t  __i;                     \
    for (__i = 0; __i < __size; __i++)  \
      if (__xp[__i] != __yp[__i])       \
        return 0;                       \
    return 1;                           \
  } while (0)


/* Extend {ptr,oldsize} to {ptr,newsize} by filling with zero limbs. */
#define MPN_ZERO_EXTEND(ptr, oldsize, newsize)          \
  do {                                                  \
    ASSERT ((newsize) >= (oldsize));                    \
    MPN_ZERO ((ptr)+(oldsize), (newsize)-(oldsize));    \
  } while (0)

/* Extend {ptr,oldsize} to {ptr,newsize} by filling with "n". */
#define MPN_FILL_EXTEND(ptr, oldsize, newsize, n)       \
  do {                                                  \
    ASSERT ((newsize) >= (oldsize));                    \
    MPN_FILL ((ptr)+(oldsize), (newsize)-(oldsize), n); \
  } while (0)

/* Copy {src,ssize} to {dst,dsize}, padded out with zero limbs if necessary. */
#define MPN_COPY_ZERO_EXTEND(dst,dsize,src,ssize)       \
  do {                                                  \
    ASSERT (dsize >= ssize);                            \
    MPN_COPY (dst, src, ssize);                         \
    MPN_ZERO_EXTEND (dst, ssize, dsize);                \
  } while (0)

#define MPN_FILL(ptr, size, n)          \
  do {                                  \
    mp_size_t __i;                      \
    ASSERT ((size) >= 0);               \
    for (__i = 0; __i < (size); __i++)  \
      (ptr)[__i] = (n);                 \
  } while (0)

#define MPN_LSHIFT_OR_COPY(wp,wsize, xp,xsize, shift)   \
  do {                                                  \
    (wsize) = (xsize);                                  \
    if ((shift) != 0)                                   \
      {                                                 \
        mp_limb_t  __c;                                 \
        __c = mpn_lshift (wp, xp, wsize, shift);        \
        wp[wsize] = __c;                                \
        (wsize) += (__c != 0);                          \
      }                                                 \
    else                                                \
      MPN_COPY (wp, xp, wsize);                         \
  } while (0)

#define MPN_LSHIFTMUL_1_EXTEND(wp,wsize, xp,xsize, y, shift)    \
  do {                                                          \
    mp_limb_t  __c;                                             \
    mp_srcptr  __xp;                                            \
    (wsize) = (xsize);                                          \
    __xp = (xp);                                                \
    if ((shift) == 0 || (y >> (GMP_NUMB_BITS - shift)) == 0)    \
      {                                                         \
        __c = mpn_mul_1 (pp, __xp, msize, (y) << shift);        \
      }                                                         \
    else                                                        \
      {                                                         \
        __c = mpn_mul_1 (pp, __xp, msize, y);                   \
        (wp)[wsize] = __c;                                      \
        (wsize) += (__c != 0);                                  \
        __c = mpn_lshift (pp, pp, msize, shift);                \
      }                                                         \
    (wp)[wsize] = __c;                                          \
    (wsize) += (__c != 0);                                      \
  } while (0)

#if WANT_TMP_DEBUG
#define TMP_ALLOC_LIMBS_2SEP(xp,xsize, ap,bp,bsize)     \
  do {                                                  \
    (xp) = TMP_ALLOC_LIMBS (xsize);                     \
    if ((ap) == (bp))                                   \
      {                                                 \
        mp_ptr  __bnewp = TMP_ALLOC_LIMBS (bsize);      \
        MPN_COPY (__bnewp, bp, bsize);                  \
        (bp) = __bnewp;                                 \
      }                                                 \
  } while (0)
#else
#define TMP_ALLOC_LIMBS_2SEP(xp,xsize, ap,bp,bsize)                     \
  do {                                                                  \
    (xp) = TMP_ALLOC_LIMBS ((xsize) + ((ap) == (bp) ? (bsize) : 0));    \
    if ((ap) == (bp))                                                   \
      {                                                                 \
        mp_ptr  __bnewp = (xp) + (xsize);                               \
        MPN_COPY (__bnewp, bp, bsize);                                  \
        (bp) = __bnewp;                                                 \
      }                                                                 \
  } while (0)
#endif

#if WANT_TMP_DEBUG
#define TMP_ALLOC_LIMBS_3(xp,xsize, yp,ysize, zp,zsize) \
  do {                                                  \
    ASSERT ((xsize) > 0);                               \
    ASSERT ((ysize) > 0);                               \
    ASSERT ((zsize) > 0);                               \
                                                        \
    (xp) = TMP_ALLOC_LIMBS (xsize);                     \
    (yp) = TMP_ALLOC_LIMBS (ysize);                     \
    (zp) = TMP_ALLOC_LIMBS (zsize);                     \
  } while (0)
#else
#define TMP_ALLOC_LIMBS_3(xp,xsize, yp,ysize, zp,zsize)         \
  do {                                                          \
    ASSERT ((xsize) > 0);                                       \
    ASSERT ((ysize) > 0);                                       \
    ASSERT ((zsize) > 0);                                       \
                                                                \
    (xp) = TMP_ALLOC_LIMBS ((xsize) + (ysize) + (zsize));       \
    (yp) = (xp) + (xsize);                                      \
    (zp) = (yp) + (ysize);                                      \
  } while (0)
#endif

#if WANT_TMP_DEBUG
#define TMP_ALLOC_LIMBS_4(wp,wsize, xp,xsize, yp,ysize, zp,zsize)       \
  do {                                                                  \
    ASSERT ((wsize) > 0);                                               \
    ASSERT ((xsize) > 0);                                               \
    ASSERT ((ysize) > 0);                                               \
    ASSERT ((zsize) > 0);                                               \
                                                                        \
    (wp) = TMP_ALLOC_LIMBS (wsize);                                     \
    (xp) = TMP_ALLOC_LIMBS (xsize);                                     \
    (yp) = TMP_ALLOC_LIMBS (ysize);                                     \
    (zp) = TMP_ALLOC_LIMBS (zsize);                                     \
  } while (0)
#else
#define TMP_ALLOC_LIMBS_4(wp,wsize, xp,xsize, yp,ysize, zp,zsize)       \
  do {                                                                  \
    ASSERT ((wsize) > 0);                                               \
    ASSERT ((xsize) > 0);                                               \
    ASSERT ((ysize) > 0);                                               \
    ASSERT ((zsize) > 0);                                               \
                                                                        \
    (wp) = TMP_ALLOC_LIMBS ((wsize) + (xsize) + (ysize) + (zsize));     \
    (xp) = (wp) + (wsize);                                              \
    (yp) = (xp) + (xsize);                                              \
    (zp) = (yp) + (ysize);                                              \
  } while (0)
#endif

#if WANT_TMP_DEBUG
#define TMP_ALLOC_LIMBS_4SEP(xp,xsize, yp,ysize, zp,zsize, ap,bp,bsize) \
  do {                                                                  \
    ASSERT ((xsize) > 0);                                               \
    ASSERT ((ysize) > 0);                                               \
    ASSERT ((zsize) > 0);                                               \
    ASSERT ((bsize) > 0);                                               \
                                                                        \
    (xp) = TMP_ALLOC_LIMBS (xsize);                                     \
    (yp) = TMP_ALLOC_LIMBS (ysize);                                     \
    (zp) = TMP_ALLOC_LIMBS (zsize);                                     \
    if ((ap) == (bp))                                                   \
      {                                                                 \
        mp_ptr  __bnewp = TMP_ALLOC_LIMBS (bsize);                      \
        MPN_COPY (__bnewp, bp, bsize);                                  \
        bp = __bnewp;                                                   \
      }                                                                 \
  } while (0)
#else
#define TMP_ALLOC_LIMBS_4SEP(xp,xsize, yp,ysize, zp,zsize, ap,bp,bsize) \
  do {                                                                  \
    ASSERT ((xsize) > 0);                                               \
    ASSERT ((ysize) > 0);                                               \
    ASSERT ((zsize) > 0);                                               \
    ASSERT ((bsize) > 0);                                               \
                                                                        \
    (xp) = TMP_ALLOC_LIMBS ((xsize) + (ysize) + (zsize)                 \
                            + ((ap) == (bp) ? (bsize) : 0));            \
    (yp) = (xp) + (xsize);                                              \
    (zp) = (yp) + (ysize);                                              \
    if ((ap) == (bp))                                                   \
      {                                                                 \
        mp_ptr  __bnewp = (zp) + (zsize);                               \
        MPN_COPY (__bnewp, bp, bsize);                                  \
        bp = __bnewp;                                                   \
      }                                                                 \
  } while (0)
#endif


/* ph:pl = low 2 limbs of xh:xl * yh:yl, no nails */
#define umul_p2m2m2(ph,pl, xh,xl, yh,yl)                \
  do {                                                  \
    mp_limb_t  __umul222__ph, __umul222__pl;            \
    umul_ppmm (__umul222__ph, __umul222__pl, xl, yl);   \
    (ph) = __umul222__ph + (xh)*(yl) + (xl)*(yh);       \
    (pl) = __umul222__pl;                               \
  } while (0)

/* p2:p1:p0 = low 3 limbs of x2:x1:x0 * y2:y1:y0 */
#define umul_p3m3m3(p2,p1,p0, x2,x1,x0, y2,y1,y0)               \
  do {                                                          \
    mp_limb_t  __umul333__p2, __umul333__p1, __umul333__p0;     \
    mp_limb_t  __umul333__mh, __umul333__ml;                    \
                                                                \
    umul_ppmm (__umul333__p1,__umul333__p0, x0, y0);            \
    __umul333__p2 = (x2)*(y0) + (x1)*(y1) + (x0)*(y1);          \
                                                                \
    umul_ppmm (__umul333__mh,__umul333__ml, x1, y0);            \
    add_ssaaaa (__umul333__p2,__umul333__p1,                    \
                __umul333__p2,__umul333__p1,                    \
                __umul333__mh,__umul333__ml);                   \
                                                                \
    umul_ppmm (__umul333__mh,__umul333__ml, x0, y1);            \
    add_ssaaaa (__umul333__p2,__umul333__p1,                    \
                __umul333__p2,__umul333__p1,                    \
                __umul333__mh,__umul333__ml);                   \
                                                                \
    (p2) = __umul333__p2;                                       \
    (p1) = __umul333__p1;                                       \
    (p0) = __umul333__p0;                                       \
  } while (0)



/* w1:w0 = x0 * y0, any nails */
#define MUL_211(w1,w0, x0, y0)                          \
  do {                                                  \
    ASSERT_LIMB (x0);                                   \
    ASSERT_LIMB (y0);                                   \
    umul_ppmm (w1, w0, x0, (y0) << GMP_NAIL_BITS);      \
    (w0) >>= GMP_NAIL_BITS;                             \
  } while (0)

/* w1:w0 = low 2 limbs of x1:x0 * y1:y0, any nails */
#define MUL_222(w1,w0, x1,x0, y1,y0)                                    \
  do {                                                                  \
    mp_limb_t  __mul222__w1, __mul222__w0;                              \
    ASSERT_LIMB (x1);                                                   \
    ASSERT_LIMB (x0);                                                   \
    ASSERT_LIMB (y1);                                                   \
    ASSERT_LIMB (y0);                                                   \
                                                                        \
    MUL_211 (__mul222__w1,__mul222__w0, x0, y0);                        \
    (w1) = (__mul222__w1 + (x1)*(y0) + (x0)*(y1)) & GMP_NUMB_MASK;      \
    (w0) = __mul222__w0;                                                \
  } while (0)

/* w2:w1:w0 = low 3 limbs of x2:x1:x0 * y2:y1:y0 */
#define MUL_333(w2,w1,w0, x2,x1,x0, y2,y1,y0)                           \
  do {                                                                  \
    mp_limb_t  __mul333__w2, __mul333__w1, __mul333__w0;                \
    mp_limb_t  __mul333__mh, __mul333__ml;                              \
    ASSERT_LIMB (x2);                                                   \
    ASSERT_LIMB (x1);                                                   \
    ASSERT_LIMB (x0);                                                   \
    ASSERT_LIMB (y2);                                                   \
    ASSERT_LIMB (y1);                                                   \
    ASSERT_LIMB (y0);                                                   \
                                                                        \
    MUL_211 (__mul333__w1,__mul333__w0, x0, y0);                        \
    __mul333__w2 = (x2)*(y0) + (x1)*(y1) + (x0)*(y1);                   \
                                                                        \
    MUL_211 (__mul333__mh,__mul333__ml, x1, y0);                        \
    ADD_2 (__mul333__w2,__mul333__w1,                                   \
           __mul333__w2,__mul333__w1, __mul333__mh,__mul333__ml);       \
                                                                        \
    MUL_211 (__mul333__mh,__mul333__ml, x0, y1);                        \
    ADD_2 (__mul333__w2,__mul333__w1,                                   \
           __mul333__w2,__mul333__w1, __mul333__mh,__mul333__ml);       \
                                                                        \
    (w2) = __mul333__w2 & GMP_NUMB_MASK;                                \
    (w1) = __mul333__w1;                                                \
    (w0) = __mul333__w0;                                                \
  } while (0)

/* w2:w1:w0 = low 3 limbs of x2:x1:x0 * y2:y1:y0 */
#define MUL_444(w3,w2,w1,w0, x3,x2,x1,x0, y3,y2,y1,y0)          \
  do {                                                          \
    mp_limb_t  __w3, __w2, __w1, __w0, __mh, __ml;              \
    ASSERT_LIMB (x3);                                           \
    ASSERT_LIMB (x2);                                           \
    ASSERT_LIMB (x1);                                           \
    ASSERT_LIMB (x0);                                           \
    ASSERT_LIMB (y3);                                           \
    ASSERT_LIMB (y2);                                           \
    ASSERT_LIMB (y1);                                           \
    ASSERT_LIMB (y0);                                           \
                                                                \
    abort ();   /* this code not working yet */                 \
                                                                \
    MUL_211 (__w1,__w0, x0, y0);                                \
    MUL_211 (__w3,__w2, x0, y2);                                \
                                                                \
    __w3 += (x3)*(y0) + (x2)*(y1) + (x1)*(y2) + (x0)*(y3);      \
                                                                \
    MUL_211 (__mh,__ml, x1, y0);                                \
    ADD_2 (__w2,__w1, __w2,__w1, __mh,__ml);                    \
                                                                \
    MUL_211 (__mh,__ml, x0, y1);                                \
    ADD_2 (__w2,__w1, __w2,__w1, __mh,__ml);                    \
                                                                \
    (w2) = __w2 & GMP_NUMB_MASK;                                \
    (w1) = __w1;                                                \
    (w0) = __w0;                                                \
  } while (0)


/* wh:wl = twos-complement negative of xh:xl, for xl non-zero */
#define NEG_2_LOWNZ(wh,wl, xh,xl)               \
  do {                                          \
    mp_limb_t  __neg2nz__xh = (xh);             \
    mp_limb_t  __neg2nz__xl = (xl);             \
    ASSERT_LIMB (xh);                           \
    ASSERT_LIMB (xl);                           \
    ASSERT ((xl) != 0);                         \
    (wl) = (-__neg2nz__xl) & GMP_NUMB_MASK;     \
    (wh) = (~__neg2nz__xh) & GMP_NUMB_MASK;     \
  } while (0)

/* w2:w1:w0 = twos-complement negative of x2:x1:x0,
   with x2:x1:x0 assumed to be not zero */
#define NEG_3_LOWNZ(w2,w1,w0, x2,x1,x0)               \
  do {                                                  \
    mp_limb_t  __neg3nz__x2 = (x2);                             \
    mp_limb_t  __neg3nz__x1 = (x1);                             \
    mp_limb_t  __neg3nz__x0 = (x0);                             \
    ASSERT_LIMB (x2);                                   \
    ASSERT_LIMB (x1);                                   \
    ASSERT_LIMB (x0);                                   \
    ASSERT ((x2) != 0 || (x1) != 0 || (x0) != 0);       \
    (w0) = (-__neg3nz__x0) & GMP_NUMB_MASK;                     \
    (w1) = (~__neg3nz__x1) & GMP_NUMB_MASK;                     \
    (w2) = (~__neg3nz__x2) & GMP_NUMB_MASK;                     \
  } while (0)

/* w1:w0 = x1:x0 + y1:y0 */
#define ADD_2(w1,w0, x1,x0, y1,y0)                      \
  do {                                                  \
    ASSERT_LIMB (x1);                                   \
    ASSERT_LIMB (x0);                                   \
    ASSERT_LIMB (y1);                                   \
    ASSERT_LIMB (y0);                                   \
                                                        \
    if (GMP_NAIL_BITS == 0)                             \
      add_ssaaaa (w1,w0, x1,x0, y1,y0);                 \
    else                                                \
      {                                                 \
        mp_limb_t   __add2__w0, __add2__c;              \
        ADDC_LIMB (__add2__c, __add2__w0, x0, y0);      \
        (w1) = (x1) + (y1) + __add2__c;                 \
        (w0) = __add2__w0;                              \
      }                                                 \
  } while (0)

/* w1:w0 = x1:x0 + y0 */
#define ADD_221(w1,w0, x1,x0, y0)                       \
  do {                                                  \
    ASSERT_LIMB (x1);                                   \
    ASSERT_LIMB (x0);                                   \
    ASSERT_LIMB (y0);                                   \
                                                        \
    if (GMP_NAIL_BITS == 0)                             \
      add_ssaaaa (w1,w0, x1,x0, CNST_LIMB(0),y0);       \
    else                                                \
      {                                                 \
        mp_limb_t  __add221__w0, __add221__c;           \
        ADDC_LIMB (__add221__c, __add221__w0, x0, y0);  \
        (w1) = ((x1) + __add221__c) & GMP_NUMB_MASK;    \
        (w0) = __add221__w0;                            \
      }                                                 \
  } while (0)

/* w1:w0 = x1:x0 + y0, with it assumed there's no carry out of w1, ie. the
   sum fits in two limbs (no-carry is only a consistency check here, it
   doesn't help speed)  */
#define ADD_221_NC(w1,w0, x1,x0, y0)                    \
  do {                                                  \
    ASSERT_CODE (mp_limb_t __add221nc__x1 = (x1));      \
    ADD_221 (w1,w0, x1,x0, y0);                         \
    ASSERT ((w1) <= GMP_NUMB_MAX);                      \
    ASSERT ((w1) >= __add221nc__x1);                    \
  } while (0)

/* w1:w0 = x1:x0 + y0 + z0, with it assumed there's no carry out of w1,
   ie. the sum fits in two limbs (no-carry is only a consistency check here,
   it doesn't help speed)  */
#if GMP_NAIL_BITS <= 1
#define ADD_2211_NC(w1,w0, x1,x0, y0, z0)                       \
  do {                                                          \
    mp_limb_t  __add2211nc__w1, __add2211nc__w0;                \
    ASSERT_LIMB (x1);                                           \
    ASSERT_LIMB (x0);                                           \
    ASSERT_LIMB (y0);                                           \
    ASSERT_LIMB (z0);                                           \
    ADD_221_NC (__add2211nc__w1,__add2211nc__w0, x1,x0, y0);    \
    ADD_221_NC (__add2211nc__w1,__add2211nc__w0,                \
                __add2211nc__w1,__add2211nc__w0, z0);           \
    (w1) = __add2211nc__w1;                                     \
    (w0) = __add2211nc__w0;                                     \
  } while (0)
#else /* nails >= 2 */
#define ADD_2211_NC(w1,w0, x1,x0, y0, z0)               \
  do {                                                  \
    mp_limb_t  __add2211nc__w0;                         \
    ASSERT_LIMB (x1);                                   \
    ASSERT_LIMB (x0);                                   \
    ASSERT_LIMB (y0);                                   \
    ASSERT_LIMB (z0);                                   \
    __add2211nc__w0 = (x0) + (y0) + (z0);               \
    (w1) = (x1) + (__add2211nc__w0 >> GMP_NUMB_BITS);   \
    (w0) = __add2211nc__w0 & GMP_NUMB_MASK;             \
    ASSERT_LIMB (w1);                                   \
  } while (0)
#endif

/* w2:w1:w0 = x2:x1:x0 + y0 */
#define ADD_331(w2,w1,w0, x2,x1,x0, y0)                         \
  do {                                                          \
    mp_limb_t  __add331__w1, __add331__w0, __add331__c;         \
    ASSERT_LIMB (x2);                                           \
    ASSERT_LIMB (x1);                                           \
    ASSERT_LIMB (x0);                                           \
    ASSERT_LIMB (y0);                                           \
                                                                \
    ADDC_LIMB (__add331__c, __add331__w0, x0, y0);              \
    ADDC_LIMB (__add331__c, __add331__w1, x1,  __add331__c);    \
    (w2) = ((x2) + __add331__c) & GMP_NUMB_MASK;                \
    (w1) = __add331__w1;                                        \
    (w0) = __add331__w0;                                        \
  } while (0)

/* w2:w1:w0 = x2:x1:x0 + y1:y0 */
#if GMP_NAIL_BITS == 0
#define ADD_332(w2,w1,w0, x2,x1,x0, y1,y0)                      \
  do {                                                          \
    mp_limb_t  __add332__w1, __add332__w0;                      \
    add_ssaaaa (__add332__w1,__add332__w0, x1,x0, y1,y0);       \
    (w2) = (x2) + (__add332__w1 < (y1));                        \
    (w1) = __add332__w1;                                        \
    (w0) = __add332__w0;                                        \
  } while (0)
#else
#define ADD_332(w2,w1,w0, x2,x1,x0, y1,y0)      \
  do {                                          \
    mp_limb_t  __add332__w1, __add332__w0, __add332__c;                 \
    ASSERT_LIMB (x2);                           \
    ASSERT_LIMB (x1);                           \
    ASSERT_LIMB (x0);                           \
    ASSERT_LIMB (y1);                           \
    ASSERT_LIMB (y0);                           \
                                                \
    ADDC_LIMB (__add332__c, __add332__w0, x0, y0);              \
    ADDC_LIMB (__add332__c, __add332__w1, x1, (y1) + __add332__c);      \
    (w2) = ((x2) + __add332__c) & GMP_NUMB_MASK;        \
    (w1) = __add332__w1;                                \
    (w0) = __add332__w0;                                \
  } while (0)
#endif

/* w1:w0 = x1:x0 - y1:y0 */
#define SUB_2(w1,w0, x1,x0, y1,y0)                              \
  do {                                                          \
    ASSERT_LIMB (x1);                                           \
    ASSERT_LIMB (x0);                                           \
    ASSERT_LIMB (y1);                                           \
    ASSERT_LIMB (y0);                                           \
                                                                \
    if (GMP_NAIL_BITS == 0)                                     \
      sub_ddmmss (w1,w0, x1,x0, y1,y0);                         \
    else                                                        \
      {                                                         \
        mp_limb_t   __sub2__w0, __sub2__c;                      \
        SUBC_LIMB (__sub2__c, __sub2__w0, x0, y0);              \
        (w1) = ((x1) - (y1) - __sub2__c) & GMP_NUMB_MASK;       \
        (w0) = __sub2__w0;                                      \
      }                                                         \
  } while (0)


/* q,r = a1:a0 / d, nails allowed */
#define UDIV_1121(q, r, n1,n0, d)                                       \
  do {                                                                  \
    mp_limb_t  __udiv1121__r;                                           \
    ASSERT_LIMB (n1);                                                   \
    ASSERT_LIMB (n0);                                                   \
    ASSERT_LIMB (d);                                                    \
    udiv_qrnnd (q, __udiv1121__r,                                       \
                n1, (n0) << GMP_NAIL_BITS, (d) << GMP_NAIL_BITS);       \
    (r) = __udiv1121__r >> GMP_NAIL_BITS;                               \
  } while (0)


/* ih set so that ih:il * nh:nl == 1 mod R^2,
   starting from il satisfying il*nl==1 mod R */
#define modlimb_invert_1to2(ih, il, nh,nl)                      \
  do {                                                          \
    mp_limb_t  __modlinv1to2__th, __modlinv1to2__tl;            \
    mp_limb_t  __modlinv1to2__il2h, __modlinv1to2__il2l;        \
    mp_limb_t  __modlinv1to2__il_new;                           \
    ASSERT_LIMB (il);                                           \
    ASSERT_LIMB (nh);                                           \
    ASSERT_LIMB (nl);                                           \
                                                                \
    MUL_211 (__modlinv1to2__th,__modlinv1to2__tl, il, il);      \
    MUL_222 (__modlinv1to2__th,__modlinv1to2__tl,               \
             __modlinv1to2__th,__modlinv1to2__tl, nh,nl);       \
    LSHIFT_21 (__modlinv1to2__il2h,__modlinv1to2__il2l, il, 1); \
    SUB_2 (ih, __modlinv1to2__il_new,                           \
           __modlinv1to2__il2h,__modlinv1to2__il2l,             \
           __modlinv1to2__th, __modlinv1to2__tl);               \
    ASSERT (__modlinv1to2__il_new == (il));                     \
  } while (0)

/* set ih:il to a value satisfying ih:il * nh:nl == 1 mod R^2 */
#define modlimb_invert_2(ih,il, nh,nl)                          \
  do {                                                          \
    mp_limb_t  __modlinv2__nl = (nl);                           \
    mp_limb_t  __modlinv2__nh = (nh);                           \
    ASSERT_LIMB (nh);                                           \
    ASSERT_LIMB (nl);                                           \
                                                                \
    modlimb_invert (il, __modlinv2__nl);                        \
    modlimb_invert_1to2 (ih,il, __modlinv2__nh,__modlinv2__nl); \
  } while (0)

/* set i2:i1:i0 to a value satisfying i2:i1:i0 * n2:n1:n0 == 1 mod R^3 */
#define modlimb_invert_3(i2,i1,i0, n2,n1,n0)                    \
  do {                                                          \
    mp_limb_t  __modlinv3__n0 = (n0);                           \
    mp_limb_t  __modlinv3__n1 = (n1);                           \
    mp_limb_t  __modlinv3__n2 = (n2);                           \
    abort ();   /* think this code is buggy ... */              \
    modlimb_invert (i0, __modlinv3__n0);                        \
    modlimb_invert_1to2 (i1,i0, __modlinv3__n1,__modlinv3__n2); \
  } while (0)

/* set inv to a value satisfying -inv * n == 1 mod R^3 */
#define modlimb_neginvert(inv, n)     \
  do {                                \
    modlimb_invert (inv, n);          \
    (inv) = -(inv);                   \
  } while (0)

/* w1:w0 = x0 << shift, with nails, must have shift>0 */
#define LSHIFT_21(w1,w0, x0, shift)                                     \
  do {                                                                  \
    mp_limb_t  __lshift21__x0 = (x0);                                   \
    int        __lshift21__shift = (shift);                             \
    ASSERT_LIMB (x0);                                                   \
    ASSERT ((shift) > 0);                                               \
    ASSERT ((shift) < GMP_NUMB_BITS);                                   \
    (w0) = (__lshift21__x0 << __lshift21__shift) & GMP_NUMB_MASK;       \
    (w1) = (__lshift21__x0 >> (GMP_NUMB_BITS - __lshift21__shift));     \
  } while (0)

/* w1:w0 = x0 << shift, with nails, with shift==0 allowed */
#define LSHIFT_21_ANY(w1,w0, x0, shift)                                 \
  do {                                                                  \
    int        __lshift21any__shift = (shift);                          \
    mp_limb_t  __lshift21any__x0 = (x0);                                \
    ASSERT_LIMB (x0);                                                   \
    ASSERT ((shift) >= 0);                                              \
    ASSERT ((shift) < GMP_NUMB_BITS);                                   \
    (w0) = __lshift21any__x0;                                           \
    (w1) = 0;                                                           \
    if (__lshift21any__shift != 0)                                      \
      LSHIFT_21 (w1, w0, __lshift21any__x0, __lshift21any__shift);      \
  } while (0)

/* w1:w0 = x1:x0 >> shift, allowing for nails */
#define LSHIFT_2(w1,w0, x1,x0, shift)                                   \
  do {                                                                  \
    mp_limb_t  __lshift2__x1 = (x1);                                    \
    mp_limb_t  __lshift2__x0 = (x0);                                    \
    int        __lshift2__shift = (shift);                              \
    ASSERT_LIMB (x1);                                                   \
    ASSERT_LIMB (x0);                                                   \
    ASSERT ((shift) > 0);                                               \
    ASSERT ((shift) < GMP_NUMB_BITS);                                   \
                                                                        \
    (w0) = (__lshift2__x0 << __lshift2__shift) & GMP_NUMB_MASK;         \
    (w1) = ((__lshift2__x1 << __lshift2__shift) & GMP_NUMB_MASK)        \
      | (__lshift2__x0 >> (GMP_NUMB_BITS - __lshift2__shift));          \
                                                                        \
  } while (0)

/* w2:w1:w0 = x2:x1:x0 << shift, discard carry, nails required */
#define LSHIFT_3(w2,w1,w0, x2,x1,x0, shift)                             \
  do {                                                                  \
    mp_limb_t  __lshift3__x2 = (x2);                                    \
    mp_limb_t  __lshift3__x1 = (x1);                                    \
    mp_limb_t  __lshift3__x0 = (x0);                                    \
    int        __lshift3__shift = (shift);                              \
    ASSERT_LIMB (x2);                                                   \
    ASSERT_LIMB (x1);                                                   \
    ASSERT_LIMB (x0);                                                   \
    ASSERT ((shift) > 0);                                               \
    ASSERT ((shift) < GMP_NUMB_BITS);                                   \
                                                                        \
    (w0) = (__lshift3__x0 << __lshift3__shift) & GMP_NUMB_MASK;         \
    (w1) = ((__lshift3__x1 << __lshift3__shift) & GMP_NUMB_MASK)        \
      | (__lshift3__x0 >> (GMP_NUMB_BITS - __lshift3__shift));          \
    (w2) = ((__lshift3__x2 << __lshift3__shift) & GMP_NUMB_MASK)        \
      | (__lshift3__x1 >> (GMP_NUMB_BITS - __lshift3__shift));          \
  } while (0)

/* w2:w1:w0 = x1:x0 << shift, nails required */
#define LSHIFT_32(w2,w1,w0, x1,x0, shift)       \
  do {                                          \
    mp_limb_t  __lshift32__x1 = (x1);                     \
    mp_limb_t  __lshift32__x0 = (x0);                     \
    int        __lshift32__shift = (shift);               \
    ASSERT_LIMB (x1);                           \
    ASSERT_LIMB (x0);                           \
    ASSERT ((shift) > 0);                       \
    ASSERT ((shift) < GMP_NUMB_BITS);           \
                                                \
    (w0) = (__lshift32__x0 << __lshift32__shift) & GMP_NUMB_MASK;   \
    (w1) = ((__lshift32__x1 << __lshift32__shift) & GMP_NUMB_MASK)  \
      | (__lshift32__x0 >> (GMP_NUMB_BITS-__lshift32__shift));      \
    (w2) = (__lshift32__x1 >> (GMP_NUMB_BITS-__lshift32__shift));   \
  } while (0)


/* w1:w0 = x1:x0 >> shift, allowing for nails */
#define RSHIFT_2(w1,w0, x1,x0, shift)                           \
  do {                                                          \
    mp_limb_t  __rshift2__x1 = (x1);                            \
    mp_limb_t  __rshift2__x0 = (x0);                            \
    int        __rshift2__shift = (shift);                      \
    ASSERT_LIMB (x1);                                           \
    ASSERT_LIMB (x0);                                           \
    ASSERT ((shift) > 0);                                       \
    ASSERT ((shift) < GMP_NUMB_BITS);                           \
                                                                \
    (w1) = (__rshift2__x1 >> __rshift2__shift);                 \
    (w0) = (__rshift2__x0 >> __rshift2__shift)                  \
      | ((__rshift2__x1 << (GMP_NUMB_BITS-__rshift2__shift))    \
         & GMP_NUMB_MASK);                                      \
                                                                \
  } while (0)




#define mpn_mullowhalf_basecase __MPN(mullowhalf_basecase)
void mpn_mullowhalf_basecase __GMP_PROTO ((mp_ptr, mp_srcptr, mp_srcptr, mp_size_t));

#define mpn_mullowhalf_kara __MPN(mullowhalf_kara)
void mpn_mullowhalf_kara __GMP_PROTO ((mp_ptr, mp_srcptr, mp_srcptr, mp_size_t));

#define mpn_mulhighhalf_basecase __MPN(mulhighhalf_basecase)
void mpn_mulhighhalf_basecase __GMP_PROTO ((mp_ptr, mp_srcptr, mp_srcptr, mp_size_t));

#define mpn_mulhighhalf_kara __MPN(mulhighhalf_kara)
void mpn_mulhighhalf_kara __GMP_PROTO ((mp_ptr, mp_srcptr, mp_srcptr, mp_size_t));

#ifndef MULLOWHALF_KARATSUBA_THRESHOLD
#define MULLOWHALF_KARATSUBA_THRESHOLD   52
#endif

#define mpn_mullowhalf_n(wp, xp, yp, size)                      \
  do {                                                          \
    if (BELOW_THRESHOLD (size, MULLOWHALF_KARATSUBA_THRESHOLD)) \
      mpn_mullowhalf_basecase (wp, xp, yp, size);               \
    else                                                        \
      mpn_mullowhalf_kara (wp, xp, yp, size);                   \
  } while (0)

/* FIXME */
#define mpn_sqrlowhalf_n(wp, xp, size)          \
  do {                                          \
    mpn_mullowhalf_n (wp, xp, xp, size);        \
  } while (0)


/* placeholder */
#define mpn_mulhighhalf_n(wp, xp, yp, size)     \
  do {                                          \
    mpn_mul_n (wp, xp, yp, size);               \
  } while (0)



/* Ensure z has at least new_alloc limb, and set zp to it's data pointer. */
#define MPZ_REALLOC_PTR(zp, z, new_alloc)       \
  do {                                          \
    mp_size_t  __new_alloc = (new_alloc);       \
    ASSERT (__new_alloc >= 0);                  \
    (zp) = PTR(z);                              \
    if (UNLIKELY (__new_alloc > ALLOC(z)))      \
      (zp) = _mpz_realloc (z, __new_alloc);     \
  } while (0)

/* {wp,size} = twos-complement negative of {xp,size}, with {xp,size} assumed
   to be non-zero */
#define MPN_NEG_N_NONZERO(wp, xp, size)         \
  do {                                          \
    ASSERT_MPN_NONZERO_P (xp, size);            \
    mpn_com_n (wp, xp, size);                   \
    MPN_INCR_U (wp, msize, CNST_LIMB(1));       \
  } while (0)


/* w = x + y + cin, and set cout = 0 or 1 for carry from that, with nails */
#if GMP_NAIL_BITS == 0
#define ADDE_LIMB(cout, w, x, y, cin)           \
  do {                                          \
    mp_limb_t  __adde__x = (x);                 \
    mp_limb_t  __adde__y = (y);                 \
    mp_limb_t  __adde__cin = (cin);             \
    mp_limb_t  __adde__w, __adde__cout;         \
    ASSERT ((cin) == 0 || (cin) == 1);          \
    __adde__w = __adde__x + __adde__y;          \
    __adde__cout = __adde__w < __adde__x;       \
    __adde__w += __adde__cin;                   \
    __adde__cout += __adde__w < __adde__cin;    \
    (w) = __adde__w;                            \
    (cout) = __adde__cout;                      \
  } while (0)
#else
#define ADDE_LIMB(cout, w, x, y, cin)           \
  do {                                          \
    mp_limb_t  __adde__w;                       \
    ASSERT_LIMB (x);                            \
    ASSERT_LIMB (y);                            \
    ASSERT ((cin) == 0 || (cin) == 1);          \
    __adde__w = (x) + (y) + (cin);              \
    (w) = __adde__w & GMP_NUMB_MASK;            \
    (cout) = __adde__w >> GMP_NUMB_BITS;        \
  } while (0)
#endif

/* w = x - y - cin, and set cout = 0 or 1 for borrow from that, with nails */
#if GMP_NAIL_BITS == 0
#define SUBE_LIMB(cout, w, x, y, cin)           \
  do {                                          \
    mp_limb_t  __sube__x = (x);                 \
    mp_limb_t  __sube__y = (y);                 \
    mp_limb_t  __sube__cin = (cin);             \
    mp_limb_t  __sube__w, __sube__cout;         \
    ASSERT ((cin) == 0 || (cin) == 1);          \
    __sube__w = __sube__x - __sube__y;          \
    __sube__cout = __sube__w > __sube__x;       \
    __sube__x = __sube__w;                      \
    __sube__w = __sube__x - __sube__cin;        \
    __sube__cout += __sube__w > __sube__x;      \
    (w) = __sube__w;                            \
    (cout) = __sube__cout;                      \
  } while (0)
#else
#define SUBE_LIMB(cout, w, x, y, cin)           \
  do {                                          \
    mp_limb_t  __sube__w;                       \
    ASSERT_LIMB (x);                            \
    ASSERT_LIMB (y);                            \
    ASSERT ((cin) == 0 || (cin) == 1);          \
                                                \
    __sube__w = (x) - (y) - (cin);              \
    (w) = __sube__w & GMP_NUMB_MASK;            \
    (cout) = __sube__w >> (GMP_LIMB_BITS - 1);  \
  } while (0)
#endif

#define MPN_ADD_1(cout, wp, xp, size, y)                \
  do {                                                  \
    ASSERT ((size) >= 1);                               \
    ASSERT (MPN_SAME_OR_SEPARATE_P (wp, xp, size)); 	\
    ASSERT_MPN (xp, size);                              \
    ASSERT_LIMB (y);                                    \
    __GMPN_ADD_1 (cout, wp, xp, size, y);               \
  } while (0)

#define MPN_SUB_1(cout, wp, xp, size, y)                \
  do {                                                  \
    ASSERT ((size) >= 1);                               \
    ASSERT (MPN_SAME_OR_SEPARATE_P (wp, xp, size)); 	\
    ASSERT_MPN (xp, size);                              \
    ASSERT_LIMB (y);                                    \
    __GMPN_SUB_1 (cout, wp, xp, size, y);               \
  } while (0)

#define MPN_ADD(cout, wp, xp, xsize, yp, ysize)                 \
  do {                                                          \
    ASSERT ((ysize) >= 0);                                      \
    ASSERT ((xsize) >= (ysize));                                \
    ASSERT (MPN_SAME_OR_SEPARATE2_P (wp, xsize, xp, xsize));    \
    ASSERT (MPN_SAME_OR_SEPARATE2_P (wp, xsize, yp, ysize));    \
    __GMPN_ADD(cout, wp, xp, xsize, yp, ysize);                 \
  } while (0)

#define MPN_SUB(cout, wp, xp, xsize, yp, ysize)                 \
  do {                                                          \
    ASSERT ((ysize) >= 0);                                      \
    ASSERT ((xsize) >= (ysize));                                \
    ASSERT (MPN_SAME_OR_SEPARATE2_P (wp, xsize, xp, xsize));    \
    ASSERT (MPN_SAME_OR_SEPARATE2_P (wp, xsize, yp, ysize));    \
    __GMPN_SUB(cout, wp, xp, xsize, yp, ysize);                 \
  } while (0)


/* {wp,size} = {xp,size} + yhi:ylo, must have size>=2 */
#define MPN_ADD_2(cout, wp, xp, size, yhi, ylo)                               \
  do {                                                                        \
    mp_limb_t  __mpnadd2__ylo = (ylo);                                        \
    mp_limb_t  __mpnadd2__yhi = (yhi);                                        \
    mp_limb_t  __mpnadd2__c;                                                  \
                                                                              \
    ASSERT ((size) >= 2);                                                     \
    ASSERT_MPN (xp, size);                                                    \
    ASSERT_LIMB (yhi);                                                        \
    ASSERT_LIMB (ylo);                                                        \
    ASSERT (MPN_SAME_OR_SEPARATE_P (wp, xp, size));                           \
                                                                              \
    ADDC_LIMB (__mpnadd2__c, (wp)[0], (xp)[0], __mpnadd2__ylo);               \
    ADDE_LIMB (__mpnadd2__c, (wp)[1], (xp)[1], __mpnadd2__yhi, __mpnadd2__c); \
    if (LIKELY ((size) > 2))                                                  \
      MPN_ADD_1 (__mpnadd2__c, (wp)+2, (xp)+2, (size)-2, __mpnadd2__c);       \
    (cout) = __mpnadd2__c;                                                    \
  } while (0)

/* {wp,size} = {xp,size} + yhi:ylo, must have size>=2 */
#define MPN_SUB_2(cout, wp, xp, size, yhi, ylo)                         \
  do {                                                                  \
    mp_limb_t  __mpnsub2__c;                                            \
                                                                        \
    ASSERT ((size) >= 2);                                               \
    ASSERT_MPN (xp, size);                                              \
    ASSERT_LIMB (yhi);                                                  \
    ASSERT_LIMB (ylo);                                                  \
    ASSERT (MPN_SAME_OR_SEPARATE_P (wp, xp, size));                     \
                                                                        \
    SUBC_LIMB (__mpnsub2__c, (wp)[0], (xp)[0], ylo);                    \
    SUBE_LIMB (__mpnsub2__c, (wp)[1], (xp)[1], yhi, __mpnsub2__c);      \
    if (LIKELY ((size) > 2))                                            \
      MPN_SUB_1 (__mpnsub2__c, (wp)+2, (xp)+2, (size)-2, __mpnsub2__c); \
    (cout) = __mpnsub2__c;                                              \
  } while (0)

/* {wp,size} = {xp,size} + ylimb<<shift, must have size>=2 */
#define MPN_ADD_1_SHIFT(cout, wp, xp, size, ylimb, shift)               \
  do {                                                                  \
    mp_limb_t  __mpnadd1sh__yhi, __mpnadd1sh__ylo;                      \
    LSHIFT_21 (__mpnadd1sh__yhi, __mpnadd1sh__ylo, ylimb, shift);       \
    MPN_ADD_2 (cout, wp, xp, size, __mpnadd1sh__yhi, __mpnadd1sh__ylo); \
  } while (0)

/* {wp,size} = {xp,size} + y2:y1:y0, must have size>=3 */
#define MPN_ADD_3(cout, wp, xp, size, y2,y1,y0)                         \
  do {                                                                  \
    mp_limb_t  __add3__y2 = (y2);                                       \
    mp_limb_t  __add3__y1 = (y1);                                       \
    mp_limb_t  __add3__y0 = (y0);                                       \
    mp_limb_t  __add3__c;                                               \
                                                                        \
    ASSERT ((size) >= 3);                                               \
    ASSERT_MPN (xp, size);                                              \
    ASSERT_LIMB (y2);                                                   \
    ASSERT_LIMB (y1);                                                   \
    ASSERT_LIMB (y0);                                                   \
    ASSERT (MPN_SAME_OR_SEPARATE_P (wp, xp, size));                     \
                                                                        \
    ADDC_LIMB (__add3__c, (wp)[0], (xp)[0], __add3__y0);                \
    ADDE_LIMB (__add3__c, (wp)[1], (xp)[1], __add3__y1, __add3__c);     \
    ADDE_LIMB (__add3__c, (wp)[2], (xp)[2], __add3__y2, __add3__c);     \
    if (LIKELY ((size) > 3))                                            \
      MPN_ADD_1 (__add3__c, (wp)+3, (xp)+3, (size)-3, __add3__c);       \
    (cout) = __add3__c;                                                 \
  } while (0)



/* r = a+b mod m, expecting 0<=a,b<m and giving 0<=r<m, with nails  */
#if GMP_NAIL_BITS == 0
#define ADDMOD_1(r,a,b,m)  addmod_saam(r,a,b,m)
#else
#define ADDMOD_1(r,a,b,m)               \
  do {                                  \
    mp_limb_t  __addmod1__a = (a);      \
    mp_limb_t  __addmod1__m = (m);      \
    mp_limb_t  __addmod1__r;            \
    ASSERT_LIMB (a);                    \
    ASSERT_LIMB (b);                    \
    ASSERT ((a) < (m));                 \
    ASSERT ((b) < (m));                 \
                                        \
    __addmod1__r = __addmod1__a + (b);  \
    if (__addmod1__r >= __addmod1__m)   \
      __addmod1__r -= __addmod1__m;     \
    (r) = __addmod1__r;                 \
  } while (0)
#endif

/* r = a-b mod m, expecting 0<=a,b<m and giving 0<=r<m, with nails  */
#if GMP_NAIL_BITS == 0
#define SUBMOD_1(r,a,b,m)  submod_dmsm(r,a,b,m)
#else
#define SUBMOD_1(r,a,b,m)                       \
  do {                                          \
    mp_limb_t  __submod1__r;                    \
    ASSERT_LIMB (a);                            \
    ASSERT_LIMB (b);                            \
    ASSERT ((a) < (m));                         \
    ASSERT ((b) < (m));                         \
    __submod1__r = (a) - (b);                   \
    if ((mp_limb_signed_t) __submod1__r < 0)    \
      __submod1__r += (m);                      \
    (r) = __submod1__r;                         \
  } while (0)
#endif

/* w = -x mod m, given 0<=x<m, and returning 0<=w<x, with nails */
#define NEGMOD_1(w, x, m)                       \
  do {                                          \
    mp_limb_t  __negmod1__x = (x);              \
    mp_limb_t  __negmod1__sub;                  \
    ASSERT ((x) < (m));                         \
    __negmod1__sub = 0;                         \
    if (LIKELY (__negmod1__x != 0))             \
      __negmod1__sub = (m);                     \
    (w) = __negmod1__sub - __negmod1__x;        \
  } while (0)



/* Return non-zero if xh:xl > yh:yl */
#define CMP_2_GT(xh,xl, yh,yl)  ((xh) > (yh) || ((xh) == (yh) && (xl) > (yl)))

/* Return non-zero if xh:xl >= yh:yl */
#define CMP_2_GE(xh,xl, yh,yl)  ((xh) > (yh) || ((xh) == (yh) && (xl) >= (yl)))


/* {wp,size+1} += {xp,size}*{yp,2}, with cout set to carry limb */
#if HAVE_NATIVE_mpn_addmul_2
#define MPN_ADDMUL_2(cout, wp, xp, size, yp)    \
  do {                                          \
    (cout) = mpn_addmul_2 (wp, xp, size, yp);   \
  } while (0)
#else
#define MPN_ADDMUL_2(cout, wp, xp, size, yp)                            \
  do {                                                                  \
    mp_ptr     __wp = (wp);                                             \
    mp_srcptr  __xp = (xp);                                             \
    mp_size_t  __size = (size);                                         \
    mp_ptr     __tp;                                                    \
    mp_srcptr  __sp;                                                    \
    mp_limb_t  __c;                                                     \
    TMP_DECL (marker);                                                  \
                                                                        \
    ASSERT (MPN_SAME_OR_SEPARATE_P (wp, xp, size));                     \
    ASSERT_MPN (wp, size);                                              \
    ASSERT_MPN (xp, size);                                              \
    ASSERT_MPN (yp, 2);                                                 \
                                                                        \
    TMP_MARK (marker);                                                  \
    __sp = __xp;                                                        \
    if ((wp) == (xp))                                                   \
      {                                                                 \
        __tp = TMP_ALLOC_LIMBS (__size);                                \
        MPN_COPY (__tp, __xp, __size);                                  \
        __sp = __tp;                                                    \
      }                                                                 \
    __c = mpn_addmul_1 (__wp, __xp, __size, (yp)[0]);                   \
    ADDC_LIMB (__c, __wp[__size], __wp[__size], __c);                   \
    (cout) = __c + mpn_addmul_1 (__wp+1, __sp, __size, (yp)[1]);        \
    TMP_FREE (marker);                                                  \
  } while (0)
#endif

/* {wp,size+1} -= {xp,size}*{yp,2}, with cout set to borrow limb */
#if HAVE_NATIVE_mpn_submul_2
#define MPN_SUBMUL_2(cout, wp, xp, size, yp)    \
  do {                                          \
    (cout) = mpn_submul_2 (wp, xp, size, yp);   \
  } while (0)
#else
#define MPN_SUBMUL_2(cout, wp, xp, size, yp)                            \
  do {                                                                  \
    mp_ptr     __wp = (wp);                                             \
    mp_srcptr  __xp = (xp);                                             \
    mp_size_t  __size = (size);                                         \
    mp_ptr     __tp;                                                    \
    mp_srcptr  __sp;                                                    \
    mp_limb_t  __c;                                                     \
    TMP_DECL (marker);                                                  \
                                                                        \
    ASSERT (MPN_SAME_OR_SEPARATE_P (wp, xp, size));                     \
    ASSERT_MPN (wp, size);                                              \
    ASSERT_MPN (xp, size);                                              \
    ASSERT_MPN (yp, 2);                                                 \
                                                                        \
    TMP_MARK (marker);                                                  \
    __sp = __xp;                                                        \
    if ((wp) == (xp))                                                   \
      {                                                                 \
        __tp = TMP_ALLOC_LIMBS (__size);                                \
        MPN_COPY (__tp, __xp, __size);                                  \
        __sp = __tp;                                                    \
      }                                                                 \
    __c = mpn_submul_1 (__wp, __xp, __size, (yp)[0]);                   \
    SUBC_LIMB (__c, __wp[__size], __wp[__size], __c);                   \
    (cout) = __c + mpn_submul_1 (__wp+1, __sp, __size, (yp)[1]);        \
    TMP_FREE (marker);                                                  \
  } while (0)
#endif

/* {wp,size+1} = {xp,size}*{yp,2}, with cout set to carry limb */
#if HAVE_NATIVE_mpn_mul_2
#define MPN_MUL_2(cout, wp, xp, size, yp)       \
  do {                                          \
    (cout) = mpn_mul_2 (wp, xp, size, yp);      \
  } while (0)
#else
#define MPN_MUL_2(cout, wp, xp, size, yp)                       \
  do {                                                          \
    mp_ptr     __wp = (wp);                                     \
    mp_srcptr  __xp = (xp);                                     \
    mp_size_t  __size = (size);                                 \
    mp_ptr     __tp;                                            \
    mp_srcptr  __sp;                                            \
    TMP_DECL (marker);                                          \
                                                                \
    ASSERT (MPN_SAME_OR_SEPARATE_P (wp, xp, size));             \
    ASSERT_MPN (wp, size);                                      \
    ASSERT_MPN (xp, size);                                      \
    ASSERT_MPN (yp, 2);                                         \
                                                                \
    TMP_MARK (marker);                                          \
    __sp = __xp;                                                \
    if ((wp) == (xp))                                           \
      {                                                         \
        __tp = TMP_ALLOC_LIMBS (__size);                        \
        MPN_COPY (__tp, __xp, __size);                          \
        __sp = __tp;                                            \
      }                                                         \
    __wp[msize] = mpn_mul_1 (__wp, __xp, __size, (yp)[0]);      \
    (cout) = mpn_addmul_1 (__wp+1, __sp, __size, (yp)[1]);      \
    TMP_FREE (marker);                                          \
  } while (0)
#endif

#define MPR_SRCPTR_SWAP(x, y)                   \
  do {                                          \
    mpr_srcptr __mpr_srcptr_swap__tmp = (x);    \
    (x) = (y);                                  \
    (y) = __mpr_srcptr_swap__tmp;               \
  } while (0)


/* {ptr,size} += 1, with size!=0 and any carry out ignored */
#define MPN_ADD_ONE_INPLACE_NOTEMPTY_IGNCARRY(ptr, size)        \
  do {                                                          \
    mp_ptr     __ptr = (ptr);                                   \
    mp_size_t  __size = (size);                                 \
    mp_limb_t  __limb;                                          \
                                                                \
    ASSERT ((size) > 0);                                        \
    ASSERT_MPN (ptr, size);                                     \
                                                                \
    for (;;)                                                    \
      {                                                         \
        __limb = *__ptr;                                        \
        __limb = (__limb + 1) & GMP_NUMB_MASK;                  \
        *__ptr++ = __limb;                                      \
        if (__limb != 0)                                        \
          break;                                                \
                                                                \
        __size--;                                               \
        if (__size == 0)                                        \
          break;                                                \
      }                                                         \
  } while (0)


#if WANT_TUNE_BUILD

#undef  MULLOWHALF_KARATSUBA_THRESHOLD
#define MULLOWHALF_KARATSUBA_THRESHOLD  mullowhalf_karatsuba_threshold
extern mp_size_t                        mullowhalf_karatsuba_threshold;

#undef  REDC_BIG_THRESHOLD
#define REDC_BIG_THRESHOLD  redc_big_threshold
extern mp_size_t            redc_big_threshold;

#undef  REDC_SETZ_THRESHOLD
#define REDC_SETZ_THRESHOLD  redc_setz_threshold
extern mp_size_t             redc_setz_threshold;

#endif
